local astrolabe = DongleStub("Astrolabe-1.0")
local event = {}
local L;

local queueFunction;
local showPopup;

local currencies = {
	384, -- Dwarf
	398, -- Draenei
	393, -- Fossil
	394, -- Night Elf
	400, -- Nerubian
	397, -- Orc
	401, -- Tol'vir
	385, -- Troll
	399, -- Vrykul
}

local function updateCurrencies()
	for i=1, #currencies do
		if tonumber(currencies[i]) then
			if GetCurrencyInfo(tonumber(currencies[i])) then
				currencies[i] = GetCurrencyInfo(tonumber(currencies[i]))
			end
		end
	end
end

-- assertTable(tbl, ...)
-- Checks that a table has a given (deep) field, without throwing an error.
local function assertTable(tbl, ...)
	if not (type(tbl) == "table") then return false end
	if ... then
		if tbl[...] then
			if select(2, ...) then
				return assertTable(tbl[...], select(2, ...))
			end
		else
			return false
		end
	end
	return true
end

-- assertFields(tbl, ...)
-- Checks that a table has all listed fields. Does not check subfields.
local function assertFields(tbl, ...)
	if not (type(tbl) == "table") then return false end
	for i=1, select("#", ...) do
		if not tbl[select(i, ...)] then return false end
	end
	return true
end

-- copyTable(tbl)
-- Returns a deep copy of a table, complete with metatable.
-- Non-table references will remain as is, and self-referencing tables will crash the client.
local function copyTable(tbl)
	if type(tbl) ~= "table" then return tbl end
	local meta = getmetatable(tbl)
	if meta then meta = copyTable(meta) end
	local r = {}
	for k,v in pairs(tbl) do
		r[copyTable(k)] = copyTable(v)
	end
	setmetatable(r, meta)
	return r
end

local db = {}
db.show = true
db.track = true
local dbmeta = {
	__tostring = function(tbl)
		if assertFields(tbl, "x", "y", "gathers", "race") then
			return format("%f:%f:%d", tbl.x, tbl.y, tbl.race)
		end
	end
}

------------------------------
-- Database version control --
------------------------------

-- Note: Release versions are integers. Beta versions are decimal points.

local version = 3
-- This is the database version number, NOT the file or addon version. It's used to determine when a
--	database update is needed.

local function defaultDB()
	return {
		show = true,
		track = true,
		version = version,
		broadcast = {
			list = {
				GUILD = false,
				WHISPER = { },
				CHANNEL = { },
			},
		},
	}
end

local function updateDB(DB)
	if not DB.version then
		return defaultDB()
	end
	if DB.version == version then
		return DB
	end
	if DB.version <= 3 then
		DB.broadcast = {
			GUILD = false,
			WHISPER = { },
			CHANNEL = { },
		}
		DB.version = version
		DB.whitelist = nil
		DB.blacklist = nil
		return DB
	end
	return defaultDB()
end

-------------------------
-- Icon object manager --
-------------------------
-- (to avoid creating frames unnecessarily)

local iconPool = {}
iconPool.numIcons = 0
local shownIcons = {}

-- getIcon()
-- Returns an unused icon from the icon pool.
-- If there are no unused icons, create a new icon and return it.
local function getIcon()
	local r
	if #iconPool > 0 then
		-- Remove the icon from the pool, add it to the display, and return it
		r = iconPool[#iconPool]
		iconPool[#iconPool] = nil
		shownIcons[r] = true
		return r
	else
		-- Create a new icon, add it to the display, and return it
		iconPool.numIcons = iconPool.numIcons + 1
		r = CreateFrame("Button", "ExcavatinatorMinimapButton" .. iconPool.numIcons, UIParent, "ExcavatinatorMinimapButtonTemplate")
		shownIcons[r] = true
		return r
	end
end

-- removeIcon(icon)
-- Removes an icon from the minimap, and returns it to the icon pool.
local function removeIcon(icon)
	if shownIcons[icon] then
		astrolabe:RemoveIconFromMinimap(icon)
		iconPool[#iconPool + 1] = icon
		shownIcons[icon] = nil
	end
end

-- removeAllIcons()
-- Removes all icons from the minimap and returns them to the icon pool
local function removeAllIcons()
	for k, v in pairs(shownIcons) do
		astrolabe:RemoveIconFromMinimap(k)
		iconPool[#iconPool + 1] = k
		shownIcons[k] = nil
	end
end

-----------------------
-- Display functions --
-----------------------

-- updateIcons()
-- Shows icons within a 250 yard radius; hides all others
local function updateIcons()
	if db.show then
		local c, z, x, y = astrolabe:GetCurrentPlayerPosition()
		if db[c] then
			if db[c][z] then
				for i=1, #db[c][z] do
					local distance = astrolabe:ComputeDistance(c, z, x, y, c, z, db[c][z][i].x, db[c][z][i].y)
					if distance < 250 then
						if not assertTable(db[c][z][i], "icon", 0) then
							local icon = getIcon()
							astrolabe:PlaceIconOnMinimap(icon, c, z, db[c][z][i].x, db[c][z][i].y)
							db[c][z][i].icon = icon
						end
					else
						removeIcon(db[c][z][i].icon)
						db[c][z][i].icon = nil
					end
				end
			end
		end
	end
end

----------------------
-- Database manager --
----------------------

-- addPosition(currency, optc, optz, optx, opty)
-- Adds a gather position to the database.
-- Positions that are within 10 yards of each other are considered the same position.
local function addPosition(currency, optc, optz, optx, opty)
	if db.track then
		-- Add or update the position
		local c, z, x, y
		if (optc and optz and optx and opty) then
			c, z, x, y = optc, optz, optx, opty
		else
			c, z, x, y = astrolabe:GetCurrentPlayerPosition()
		end
		if not db[c] then db[c] = {} end -- If no entry for the current continent, create one
		if not db[c][z] then db[c][z] = {} end -- If no entry for the current zone, create one
		local found = 0
		for i=1, #db[c][z] do
			local distance = astrolabe:ComputeDistance(c, z, x, y, c, z, db[c][z][i].x, db[c][z][i].y)
			if distance < 10 then
				-- It's an old position; add one artifact found
				db[c][z][i].gathers = db[c][z][i].gathers + 1
				if currency > 0 then db[c][z][i].race = currency end -- In case of importing, don't overwrite the race
				found = i
				break
			end
		end
		if found == 0 then
			-- It's a new position; add it to the database and make an icon for the position (if it's in the zone the player is actually in)
			db[c][z][#db[c][z] + 1] = { x = x, y = y, gathers = 1, race = currency }
			local cr, zr = astrolabe:GetCurrentPlayerPosition()
			if c == cr and z == zr then
				local icon = getIcon()
				astrolabe:PlaceIconOnMinimap(icon, c, z, x, y)
				db[c][z][#db[c][z]].icon = icon
			end
		end
	end
end

-- removePosition()
-- Removes the position the player is currently at from the database.
-- Serves no purpose in the addon proper; retained for testing purposes.
local function removePosition()
	local c, z, x, y = astrolabe:GetCurrentPlayerPosition()
	for i=1, #db[c][z] do
		local distance = astrolabe:ComputeDistance(c, z, x, y, c, z, db[c][z][i].x, db[c][z][i].y)
		if distance < 10 then
			table.remove(db[c][z], i)
		end
	end
end

local function importFromArchy()
	if not ArchyDB then
		print("Excavatinator: Cannot load from Archy: No Archy database found.")
	else
		if assertTable(ArchyDB, "global", "surveyNodes") then
			for k, v in pairs(ArchyDB.global.surveyNodes) do
				for i=1, #v do
					addPosition(0, v[i].m, v[i].f, v[i].x, v[i].y)
				end
			end
			removeAllIcons()
			updateIcons()
		end
	end
end

local function resetDB()
	local show, track = db.show, db.track
	table.wipe(db)
	db.show = show
	db.track = track
	removeAllIcons()
end

-------------------
-- Data transfer --
-------------------

local progressFrame = CreateFrame("Frame", "ExcavatinatorTransferProgressFrame", UIParent)
progressFrame:SetBackdrop( {
	bgFile = "Interface\\TutorialFrame\\TutorialFrameBackground",
	edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
	tile = true, tileSize = 16, edgeSize = 16,
	insets = { left = 5, right = 5, top = 5, bottom = 5 } } )
progressFrame:SetWidth(200)
progressFrame:SetHeight(80)
progressFrame:SetPoint("CENTER")
progressFrame.header = progressFrame:CreateFontString()
progressFrame.header:SetFont("Fonts\\FRIZQT__.TTF", 12)
progressFrame.header:SetText("Sending data to someone...")
progressFrame.header:SetPoint("TOP", 0, -12)
progressFrame.footer = progressFrame:CreateFontString()
progressFrame.footer:SetFont("Fonts\\FRIZQT__.TTF", 12)
progressFrame.footer:SetText("Waiting for response")
progressFrame.footer:SetPoint("BOTTOM", 0, 12)
progressFrame.bg = CreateFrame("Frame", "ExcavatinatorTransferProgressBar", progressFrame)
progressFrame.bg:SetWidth(120)
progressFrame.bg:SetHeight(16)
progressFrame.bg:SetPoint("TOP", 0, -32)
progressFrame.bg.texture = progressFrame.bg:CreateTexture()
progressFrame.bg.texture:SetTexture(0.3, 0.3, 0.3)
progressFrame.bg.texture:SetAllPoints(progressFrame.bg)
progressFrame.progress = CreateFrame("Frame", "ExcavatinatorTransferProgressBar", progressFrame.bg)
progressFrame.progress:SetWidth(120)
progressFrame.progress:SetHeight(8)
progressFrame.progress:SetPoint("TOPLEFT", 0, 0)
progressFrame.progress.texture = progressFrame.progress:CreateTexture()
progressFrame.progress.texture:SetTexture(0, 1, 0)
progressFrame.progress.texture:SetAllPoints(progressFrame.progress)
progressFrame.progress.max = 1
progressFrame.timeout = CreateFrame("Frame", "ExcavatinatorTransferProgressBar", progressFrame.bg)
progressFrame.timeout:SetWidth(120)
progressFrame.timeout:SetHeight(8)
progressFrame.timeout:SetPoint("BOTTOMLEFT", 0, 0)
progressFrame.timeout.texture = progressFrame.timeout:CreateTexture()
progressFrame.timeout.texture:SetTexture(0, 0, 1)
progressFrame.timeout.texture:SetAllPoints(progressFrame.timeout)
progressFrame.timeout.max = 1
progressFrame:Hide()

local currentTransfer = {
	type = "none", -- "none", "send", "get", "offer", "request", "dialog" (waiting for player response)
	player = "", -- The player with whom communication is open
	time = 0, -- The timeout timer; starts at timeout length and counts down.
	data = {},
	progress = 0,
}

local DATA_PER_SECOND = 50
local TIMEOUT_DIALOG = 60
local TIMEOUT_COMM = 30

-- Message types:
-- BUSY: Response to a request or offer, when the other player is already in a transfer
-- DENY: The player has denied the data request/offer
-- REQUEST: A request for data sharing
-- OFFER: An offer to share data
-- ACCEPT_OFFER: Accepting an offer
-- DATA: Self-explanatory
-- END: End of a data transmission.

-- Data string structure:
-- "START"			Start point of the data
-- "NODE:<c>:<z>:<x>:<y>:<r>	Data for a node - no gathers data is transferred.

local function parseData(data)
	if string.sub(data[1], 1, 5) ~= "START" then
		-- Data error
	else
		local type, c, z, x, y, r = "", 0, 0, 0, 0, 0
		for i=1, #data do
			type, c, z, x, y, r = strsplit(":", data[i])
			c, z, x, y, r = tonumber(c), tonumber(z), tonumber(x), tonumber(y), tonumber(r)
			if type == "START" then
				-- c is now the amount of data to be sent.
			elseif type == "NODE" then
				if c > -1 and z > -1 then
					addPosition(r, c, z, x, y)
				else
					-- Data error
				end
			end
		end
	end
	table.wipe(currentTransfer.data)
	progressFrame:Hide()
end

local function sendDataStep(n)
	for i=n, min(n+DATA_PER_SECOND-1, #currentTransfer.data) do
		SendAddonMessage("Excavatinator_TR", currentTransfer.data[i], "WHISPER", currentTransfer.player)
		currentTransfer.progress = currentTransfer.progress + 1
		progressFrame.footer:SetText(format("%d / %d", currentTransfer.progress, progressFrame.progress.max))
	end
	if n+DATA_PER_SECOND <= #currentTransfer.data then
		queueFunction(1, sendDataStep, n+DATA_PER_SECOND)
	else
		queueFunction(2, SendAddonMessage, "Excavatinator_TR", "END", "WHISPER", currentTransfer.player)
		currentTransfer.type = "none"
		currentTransfer.player = ""
		currentTransfer.time = 0
		currentTransfer.progress = 0
		progressFrame.timeout.max = 1
		progressFrame:Hide()
		table.wipe(currentTransfer.data)
	end
end

local function sendData(player)
	if currentTransfer.type == "send" and currentTransfer.player == player then
		progressFrame:Show()
		progressFrame.header:SetText(format(L["Sending data to %s"], player))
		currentTransfer.time = 0
		progressFrame.timeout.max = 1
		table.wipe(currentTransfer.data)
		currentTransfer.data[1] = "DATASTART" -- Will be overwritten further down, but there needs to be a first value
		for c, a in pairs(db) do
			if type(a) == "table" and type(c) == "number" then
				for z,v in pairs(a) do
					for k, b in pairs(v) do
						currentTransfer.data[#currentTransfer.data + 1] = format("DATANODE:%d:%d:%s", c, z, tostring(b))
					end
				end
			end
		end
		currentTransfer.data[1] = format("DATASTART:%d:%d", #currentTransfer.data, version)
		progressFrame.progress.max = #currentTransfer.data
		sendDataStep(1)
	end
end

local function sendRequest(player)
	if not player then player = currentTransfer.player end
	if currentTransfer.type == "none" then
		progressFrame:Show()
		progressFrame.header:SetText(format(L["Request sent to %s"], player))
		progressFrame.footer:SetText("")
		currentTransfer.type = "get"
		currentTransfer.player = player
		currentTransfer.time = TIMEOUT_DIALOG -- One minute to accept.
		progressFrame.timeout.max = TIMEOUT_DIALOG
		progressFrame.timeout:Show()
		SendAddonMessage("Excavatinator_TR", "REQUEST", "WHISPER", player)
	end
end

local function sendOffer(player)
	if not player then player = currentTransfer.player end
	if currentTransfer.type == "none" then
		progressFrame:Show()
		progressFrame.header:SetText(format(L["Offer sent to %s"], player))
		progressFrame.footer:SetText("")
		currentTransfer.type = "offer"
		currentTransfer.player = player
		currentTransfer.time = TIMEOUT_DIALOG -- One minute to accept.
		progressFrame.timeout.max = TIMEOUT_DIALOG
		progressFrame.timeout:Show()
		SendAddonMessage("Excavatinator_TR", "OFFER", "WHISPER", player)
	end
end

local function acceptRequest()
	if currentTransfer.type == "dialog" then
		currentTransfer.type = "send"
		currentTransfer.time = TIMEOUT_COMM
		progressFrame.timeout.max = TIMEOUT_COMM
		progressFrame.timeout:Show()
		sendData(currentTransfer.player)
	end
end

local function acceptOffer()
	if currentTransfer.type == "dialog" then
		currentTransfer.type = "get"
		currentTransfer.time = TIMEOUT_COMM
		progressFrame.timeout.max = TIMEOUT_COMM
		progressFrame.timeout:Show()
		SendAddonMessage("Excavatinator_TR", "ACCEPT_OFFER", "WHISPER", currentTransfer.player)
	end
end

local function denyTransfer()
	if currentTransfer.type == "dialog" then
		currentTransfer.type = "none"
		currentTransfer.time = 0
		SendAddonMessage("Excavatinator_TR", "DENY", "WHISPER", currentTransfer.player)
		currentTransfer.player = ""
		currentTransfer.progress = 0
		progressFrame:Hide()
	end
end

local function cancel()
	currentTransfer.player = ""
	currentTransfer.type = "none"
end

local function timeout()
	progressFrame:Hide()
	showPopup("EXCAVATINATOR_TRANSFER_TIMEOUT")
end

local function broadcast(race)
	local c, z, x, y = astrolabe:GetCurrentPlayerPosition()
	for k, v in pairs(db.broadcast) do
		if k == "CHANNEL" then
			for a,b in pairs(v) do
				-- TODO: Channel support
			end
		elseif k == "WHISPER" then
			for a,b in pairs(v) do
				SendAddonMessage("Excavatinator_TR", format("BROADCAST%d:%d:%f:%f:%d", c, z, x, y, race), k, a)
			end
		elseif k == "GUILD" then
			if v then
				SendAddonMessage("Excavatinator_TR", format("BROADCAST%d:%d:%f:%f:%d", c, z, x, y, race), k)
			end
		end
	end
end

function event.CHAT_MSG_ADDON(prefix, msg, msgtype, player)
	if prefix == "Excavatinator_TR" then
		-- It's a transfer message.
		if msg == "BUSY" and player == currentTransfer.player then
			-- It's a BUSY response from the other player. Cancel communication.
			currentTransfer.type = "none"
			currentTransfer.player = ""
			currentTransfer.time = 0
			currentTransfer.progress = 0
			table.wipe(currentTransfer.data)
			showPopup("EXCAVATINATOR_TRANSFER_BUSY", player)
			progressFrame:Hide()
		elseif msg == "DENY" and player == currentTransfer.player then
			-- The other player has denied the request/offer.
			currentTransfer.type = "none"
			currentTransfer.player = ""
			currentTransfer.time = 0
			currentTransfer.progress = 0
			table.wipe(currentTransfer.data)
			showPopup("EXCAVATINATOR_TRANSFER_DENIED", player)
			progressFrame:Hide()
		elseif msg == "REQUEST" then
			if currentTransfer.type ~= "none" then
				-- We're busy, stop communications
				SendAddonMessage("Excavatinator_TR", "BUSY", "WHISPER", player)
			else
				-- We're ready to communicate! Ask for player confirmation.
				showPopup("EXCAVATINATOR_GET_REQUEST", player)
				currentTransfer.type = "dialog"
				currentTransfer.time = TIMEOUT_DIALOG
				progressFrame.timeout.max = TIMEOUT_DIALOG
				progressFrame.timeout:Show()
				currentTransfer.player = player
				progressFrame:Show()
				progressFrame.header:SetText(format(L["Request from %s"], player))
				progressFrame.footer:SetText("")
			end
		elseif msg == "OFFER" then
			if currentTransfer.type ~= "none" then
				-- We're busy, stop communications
				SendAddonMessage("Excavatinator_TR", "BUSY", "WHISPER", player)
			else
				-- We're ready to communicate! Ask for player confirmation.
				showPopup("EXCAVATINATOR_GET_OFFER", player)
				currentTransfer.type = "dialog"
				currentTransfer.time = TIMEOUT_DIALOG
				progressFrame.timeout.max = TIMEOUT_DIALOG
				progressFrame.timeout:Show()
				currentTransfer.player = player
				progressFrame:Show()
				progressFrame.header:SetText(format(L["Offer from %s"], player))
				progressFrame.footer:SetText("")
			end
		elseif msg == "ACCEPT_OFFER" then
			if currentTransfer.type == "offer" and currentTransfer.player == player then
				-- Our offer has been accepted, by the correct player
				currentTransfer.type = "send"
				currentTransfer.time = TIMEOUT_DIALOG
				progressFrame.timeout.max = TIMEOUT_DIALOG
				progressFrame.timeout:Show()
				sendData(currentTransfer.player)
			end
		elseif string.sub(msg, 1, 4) == "DATA" then
			if currentTransfer.type == "get" and currentTransfer.player == player then
				-- It's a data transfer. Record it.
				currentTransfer.data[#currentTransfer.data + 1] = string.sub(msg, 5)
				currentTransfer.time = TIMEOUT_COMM -- Timeout is 30 seconds after the last message.
				progressFrame.timeout.max = TIMEOUT_COMM
				progressFrame.timeout:Show()
				currentTransfer.progress = currentTransfer.progress + 1
				progressFrame.header:SetText(format(L["Getting data from %s"], player))
				progressFrame.footer:SetText(format("%d / %d", currentTransfer.progress, progressFrame.progress.max))
				progressFrame.progress:SetWidth((currentTransfer.progress / progressFrame.progress.max) * 120)
				progressFrame.progress:Show()
				if string.sub(msg, 5, 9) == "START" then
					local _, num = strsplit(":", msg)
					num = tonumber(num)
					progressFrame.progress.max = num
				end
			end
		elseif msg == "END" then
			if currentTransfer.type == "get" and currentTransfer.player == player then
				-- End of data transfer, parse the data
				parseData(currentTransfer.data)
				currentTransfer.type = "none"
				currentTransfer.player = ""
				currentTransfer.time = 0
				currentTransfer.progress = 0
			end
		elseif string.sub(msg, 1, 9) == "BROADCAST" then
			if db.broadcast[msgtype] then
				if type(db.broadcast[msgtype]) == "table" then
					if db.broadcast[msgtype][player] then
						local c, z, x, y, race = strsplit(":", string.sub(msg, 10))
						addPosition(tonumber(race), tonumber(c), tonumber(z), tonumber(x), tonumber(y))
					end
				else
					local c, z, x, y, race = strsplit(":", string.sub(msg, 10))
					addPosition(tonumber(race), tonumber(c), tonumber(z), tonumber(x), tonumber(y))
				end
			end
		end
	end
end

-------------------------
-- Confirmation frames --
-------------------------

-- Declared local
function showPopup(popup, ...)
	local oldText = StaticPopupDialogs[popup].text
	StaticPopupDialogs[popup].text = format(oldText, ...)
	StaticPopup_Show(popup)
	StaticPopupDialogs[popup].text = oldText
end

local popups = {
	["EXCAVATINATOR_DELETEDB"] = {
		text = "Are you sure you want to clear your node database? This can NOT be undone.",
		button1 = "Yes",
		button2 = "No",
		OnAccept = resetDB,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
	},
	["EXCAVATINATOR_IMPORT_ARCHY"] = {
		text = "Import node data from Archy?",
		button1 = "Yes",
		button2 = "No",
		OnAccept = importFromArchy,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
	},
	["EXCAVATINATOR_GET_REQUEST"] = {
		text = "%s is requesting that you share your node database with them. Accept?",
		button1 = "Yes",
		button2 = "No",
		OnAccept = acceptRequest,
		OnCancel = denyTransfer,
		timeout = TIMEOUT_DIALOG,
		whileDead = true,
		hideOnEscape = true,
	},
	["EXCAVATINATOR_GET_OFFER"] = {
		text = "%s is offering to share their node database. Accept?",
		button1 = "Yes",
		button2 = "No",
		OnAccept = acceptOffer,
		OnCancel = denyTransfer,
		timeout = TIMEOUT_DIALOG,
		whileDead = true,
		hideOnEscape = true,
	},
	["EXCAVATINATOR_YOU_ARE_BUSY"] = {
		text = "You are busy.",
		button1 = "OK",
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
	},
	["EXCAVATINATOR_SEND_REQUEST"] = {
		text = "Send a database transfer request to %s?",
		button1 = "Yes",
		button2 = "No",
		OnAccept = function() sendRequest(currentTransfer.player) end,
		OnCancel = cancel,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
	},
	["EXCAVATINATOR_SEND_OFFER"] = {
		text = "Offer to share your database with %s?",
		button1 = "Yes",
		button2 = "No",
		OnAccept = function() sendOffer(currentTransfer.player) end,
		OnCancel = cancel,
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
	},
	["EXCAVATINATOR_TRANSFER_DENIED"] = {
		text = "%s has denied your request.",
		button1 = "OK",
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
	},
	["EXCAVATINATOR_TRANSFER_BUSY"] = {
		text = "%s is busy.",
		button1 = "OK",
		timeout = 0,
		whileDead = true,
		hideOnEscape = true,
	},
	["EXCAVATINATOR_TRANSFER_TIMEOUT"] = {
		text = "Your transfer has timed out.",
		button1 = "OK",
		timeout = TIMEOUT_DIALOG,
		whileDead = true,
		hideOnEscape = true,
	},
}

---------------------
-- Event functions --
---------------------

function event.ADDON_LOADED(addonName)
	if addonName == "Excavatinator" then
		if ExcavatinatorNodeDB then
			db = updateDB(ExcavatinatorNodeDB)
			for k, v in pairs(db) do
				if k ~= "broadcast" and type(v) == "table" then
					for kk, vv in pairs(v) do
						for i=1, #vv do
							setmetatable(vv[i], dbmeta)
						end
					end
				end
			end
		end
		L = Excavatinator:GetLocale()
		for k,v in pairs(popups) do
			v.text = L[v.text]
			v.button1 = L[v.button1]
			v.button2 = L[v.button2]
			StaticPopupDialogs[k] = v
		end
	end
end

function event.PLAYER_LOGIN()
	if type(currencies[1]) == "number" then
		for i=1, #currencies do
			if GetCurrencyInfo(currencies[i]) == "" then
				currencies[i] = tostring(currencies[i]) -- Prevent tracking of unknown currencies.
			else
				currencies[i] = GetCurrencyInfo(currencies[i])
			end
		end
	end
	RegisterAddonMessagePrefix("Excavatinator_TR")
end

function event.PLAYER_LEAVING_WORLD()
	ExcavatinatorNodeDB = db
end

function event.CHAT_MSG_CURRENCY(msg)
	updateCurrencies()
	for i=1, #currencies do
		if strfind(msg, currencies[i]) and currencies[i] ~= "" and not tonumber(currencies[i]) then
			-- SOMETHING keeps noting 9 vrykul (race 9) finds when other currencies are gained,
			-- I have NO idea what's causing it, but somehow currencies[i] has to equal an
			-- empty string for it to happen. Again, I have NO idea how, but I'm hoping this
			-- little hack will end that problem. Like all the other little hacks I hoped would
			-- end that problem...
			addPosition(i)
			broadcast(i)
		end
	end
end

function event.ZONE_CHANGED_NEW_AREA()
	-- If the map is up when changing zones, icons may not be removed properly.
	-- Fix this by removing all icons when the zone is changed.
	removeAllIcons()
end

-------------------
-- Event manager --
-------------------

local function onEvent(s, e, ...)
	if event[e] then
		event[e](...)
	end
end

local eventFrame = CreateFrame("Frame")
eventFrame:SetScript("OnEvent", onEvent)
for k, v in pairs(event) do
	eventFrame:RegisterEvent(k)
end

----------------------
-- OnUpdate manager --
----------------------

local passed, interval, queue = 0, 1, {}

-- Declared local
function queueFunction(time, func, ...)
	if type(func) == "function" then
		queue[#queue + 1] = {}
		queue[#queue].func = func
		queue[#queue].args = { ... }
		queue[#queue].time = time
	else
	end
end

local function onUpdate(_, elapsed)
	passed = passed + elapsed
	if passed > interval then
		updateIcons()

		if currentTransfer.time > 0 then
			currentTransfer.time = currentTransfer.time - 1
			if currentTransfer.time == 0 then
				currentTransfer.type = "none"
				currentTransfer.player = ""
				currentTransfer.progress = 0
				timeout()
			end
		end

		for i=1, #queue do
			if queue[i].time then
				queue[i].time = queue[i].time - 1
				if queue[i].time <= 0 then
					queue[i].func(unpack(queue[i].args))
					table.wipe(queue[i])
				end
			end
		end

		if progressFrame:IsShown() then
			if currentTransfer.progress > 0 then
				progressFrame.progress:SetWidth((currentTransfer.progress / progressFrame.progress.max) * 120)
				progressFrame.progress:Show()
			else
				progressFrame.progress:Hide()
			end
			if currentTransfer.time > 0 then
				progressFrame.timeout:SetWidth((currentTransfer.time / progressFrame.timeout.max) * 120)
				progressFrame.timeout:Show()
			else
				progressFrame.timeout:Hide()
			end
		end

		passed = 0
	end
end

local updateFrame = CreateFrame("Frame")
updateFrame:SetScript("OnUpdate", onUpdate)

------------------------------------
-- Communicate with the main file --
------------------------------------
-- (this also adds to the API)

-- Excavatinator:SetMinimapDisplay(isOn)
-- Turns minimap icon display on or off
function Excavatinator:SetMinimapDisplay(isOn)
	db.show = isOn
	if not isOn then removeAllIcons() end
	updateIcons()
end

-- Excavatinator:GetMinimapDisplay()
-- Returns whether minimap display is on or not
function Excavatinator:GetMinimapDisplay()
	return db.show
end

-- Excavatinator:SetMinimapTracking(isOn)
-- Turns dig tracking on or off
function Excavatinator:SetMinimapTracking(isOn)
	db.track = isOn
end

-- Excavatinator:GetMinimapTracking()
-- Returns whether minimap tracking is on or off
function Excavatinator:GetMinimapTracking()
	return db.track
end

-- Excavatinator:ResetDB()
-- Resets Excavatinator's node database. The user will have to confirm this.
function Excavatinator:ResetDB()
	StaticPopup_Show("EXCAVATINATOR_DELETEDB")
end

-- Excavatinator:ImportDatabase(addonID)
-- Imports the database of a supported addon.
function Excavatinator:ImportDatabase(addonID)
	if addonID == 1 then
		StaticPopup_Show("EXCAVATINATOR_IMPORT_ARCHY")
	end
end

function Excavatinator:SendRequest(player)
	if currentTransfer.player == "" and currentTransfer.type == "none" then
		currentTransfer.player = player
		showPopup("EXCAVATINATOR_SEND_REQUEST", player)
	else
		showPopup("EXCAVATINATOR_YOU_ARE_BUSY")
	end
end

function Excavatinator:SendOffer(player)
	if currentTransfer.player == "" and currentTransfer.type == "none" then
		currentTransfer.player = player
		showPopup("EXCAVATINATOR_SEND_OFFER", player)
	else
		showPopup("EXCAVATINATOR_YOU_ARE_BUSY")
	end
end

function Excavatinator:SetBroadcast(type, player, doBroadcast)
	type = strupper(type)
	if player then player = strupper(strsub(player, 1, 1)) .. strlower(strsub(player, 2)) end
	if type == "GUILD" then
		db.broadcast.GUILD = doBroadcast
	elseif db.broadcast[type] then
		db.broadcast[type][player] = doBroadcast or nil
	end
end

function Excavatinator:GetBroadcastList()
	return copyTable(db.broadcast)
end